import os
import random

import cocos
import pyglet
from pyglet.window import key
from cocos.layer import Layer, ColorLayer, ScrollableLayer, ScrollingManager
from cocos.text import Label
from cocos.actions import ScaleBy, Repeat, Reverse, FadeOut
from cocos.scenes.transitions import RotoZoomTransition, JumpZoomTransition
from cocos.scene import Scene
from cocos.menu import *
from characters import Player, Enemy1, Enemy2, King

#Specify the resources paths for pyglet.resource
pyglet.resource.path.extend(['data', 'data/images', 'data/sounds', 'data/music', 'data/fonts'])
pyglet.resource.reindex()

#Add the data/fonts directory to the pyglet fonts directories
pyglet.font.add_directory(os.path.join(os.path.dirname(__file__), '../data/fonts'))

audio_player = pyglet.media.Player()

##################################################################################################

class PresentationLayer(Layer):
    """ Presentation layer. Very first screen of the game."""
    is_event_handler = True             #enable pyglet's events

    def __init__(self):
        super(PresentationLayer, self).__init__()
        label = Label('Himalia', font_size=40, font_name='4990810', anchor_x='center', anchor_y='center')
        label.position = 400, 300
        self.add(label)
        scale = ScaleBy(1.5, duration=3)
        scale = Repeat(scale + Reverse(scale))
        label.do(scale)

        label2 = Label('Press Enter', font_size=14, font_name='Arial Black', anchor_x='center', anchor_y='center')
        label2.position = 400, 50
        self.add(label2)
        scale = ScaleBy(1.5, duration=0.3)
        scale = Repeat(scale + Reverse(scale))
        label2.do(scale)


    def on_key_press (self, kp, modifiers):
        if kp == key.ENTER:
            color_layer = ColorLayer(0,0,0,255)
            cocos.director.director.replace(RotoZoomTransition(Scene(MainMenu()), duration=1))


##################################################################################################


class CreditsLayer(Layer):
    def __init__(self):
        super(CreditsLayer, self).__init__()
        label = Label('Credits', font_size=50, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 550
        self.add(label)

        label = Label('Programming: marciano', font_size=24, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 300
        self.add(label)

        label = Label('Music & GFX: see credits in readme.txt', font_size=24, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 250
        self.add(label)

##################################################################################################

class VictoryLayer(Layer):
    def __init__(self):
        super(VictoryLayer, self).__init__()

        if pyglet.media.have_avbin:
            audio_player.pause()
            audio_player.queue(pyglet.resource.media('Enthusiast_Audience.mp3'))
            audio_player.next()
            audio_player.play()

        label = Label('Yeah! You saved the King!', font_size=35, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 400
        self.add(label)

        label = Label('Now go for a salary increase.', font_size=25, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 250
        self.add(label)

##################################################################################################

class DefeatLayer(Layer):
    def __init__(self):
        super(DefeatLayer, self).__init__()

        if pyglet.media.have_avbin:
            audio_player.pause()
            audio_player.queue(pyglet.resource.media('Sad_Trombone.mp3'))
            audio_player.next()
            audio_player.play()

        label = Label("You couldn't save the King.", font_size=35, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 400
        self.add(label)

        label = Label("Anyways, you never supported monarchy.", font_size=25, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 250
        self.add(label)

##################################################################################################

class DifficultyLayer(Layer):
    def __init__(self):
        super(DifficultyLayer, self).__init__()
        label = Label('Choose your level', font_size=50, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 550
        self.add(label)


##################################################################################################

class HelpLayer(Layer):

    def __init__(self):
        super(HelpLayer, self).__init__()
        label = Label('Help', font_size=50, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810')
        label.position = 400, 550
        self.add(label)

        label = Label('Go rescue the King. Get close to him, and he will follow you.', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 450
        self.add(label)

        label = Label('Then lead the King to the safe place, marked with a red cross.', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 400
        self.add(label)

        label = Label('Avoid the field of view of the enemies at all costs.', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 350
        self.add(label)

        label = Label('Move with the arrow keys', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 250
        self.add(label)

        label = Label('Spacebar: Release the king (sometimes may be useful...)', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 200
        self.add(label)

        label = Label('M: Toggle music ON/OFF', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 150
        self.add(label)

        label = Label('+, -, *: Adjust zoom (only available on easy level)', font_size=16, anchor_x='center', anchor_y='center',
                      color=(255,255,255,255), font_name='4990810', halign='left')
        label.position = 400, 100
        self.add(label)


##################################################################################################

class BackgroundLayer(Layer):

    def __init__(self, img, world_width, world_height):
        super(BackgroundLayer, self).__init__()
        self.img = pyglet.resource.image(img)

    def draw(self):
        self.img.blit(0, 0)
    
##################################################################################################

class TerrainLayer(ScrollableLayer):
    is_event_handler = True

    def __init__(self, background, world_width, world_height, difficulty):
        super(TerrainLayer, self).__init__()

        #Size of the world
        self.world_width = world_width
        self.world_height= world_height
        self.add(BackgroundLayer(background, world_width, world_height), z=0)

        self.difficulty = difficulty

        #Size of the game window
        self.window_width, self.window_height = cocos.director.director.get_window_size()

        self.pressed_keys = set()

        #Put some objects and enemies in the map
        self.generate_footsteps(20)
        self.generate_houses(10)
        self.generate_vines(15)
        self.generate_pines(15)
        self.generate_rocks(15)
        number_of_enemies = {'easy': 10, 'medium': 20, 'hard':30}
        self.generate_enemies(number_of_enemies[self.difficulty])

        #Main character
        self.player = Player(self.world_width, self.world_height)
        self.add(self.player, z=1)

        #The King!
        self.king = King(self.world_width, self.world_height)
        self.add(self.king, z=1)

        if hasattr(self.player, 'shadow'):
            self.add(self.player.shadow)

        self.the_house = cocos.sprite.Sprite('house1.png')
        self.the_house.position = (self.player.position[0], self.player.position[1] + self.player.height + 90)
        self.add(self.the_house, z=0)

        #The cross, aka the safe place
        self.the_cross = cocos.sprite.Sprite('cross.png')
        self.the_cross.position = (self.the_house.position[0], self.the_house.position[1] - self.the_house.height/2 - 20)
        self.add(self.the_cross, z=0)

        safe_place = Label("This is the safe place.\nTake the King here.", font_size=14, font_name='Arial Black', color=(0, 0, 0, 255), anchor_x='center', anchor_y='center', multiline=True, width=350)
        safe_place.position = (self.the_cross.position[0] + self.the_cross.width/2 + 50, self.the_cross.position[1])
        self.add(safe_place, z=2)
        safe_place.do(FadeOut(8))


        #Play the main music theme
        if pyglet.media.have_avbin:
            audio_player.queue(pyglet.resource.media('DoKashiteru_-_The_Annual_New_England_Xylophone_Symposium.mp3', streaming=True))
            audio_player.play()

        self.schedule(self.step)

    def generate_enemies(self, quantity):
        #print "Generating %d enemies..." % quantity
        self.enemies = []
        for i in range(quantity):
            tipo = random.choice([1, 2])
            if tipo == 1:
                enemy = Enemy1(self.world_width, self.world_height)
            elif tipo == 2:
                enemy = Enemy2(self.world_width, self.world_height)
            self.enemies.append(enemy)
            self.add(enemy, z=1)                 #Add the enemy to the layer
            if hasattr(enemy, 'radar'):
                self.add(enemy.radar, z=0)

    def generate_houses(self, quantity):
        #print "Generating %d houses..." % quantity
        self.houses = []
        for i in range(quantity):
            house = cocos.sprite.Sprite('house1.png')
            house.position = (random.randint(0 + house.width/2, self.world_width-house.width-1), random.randint(200 + house.height/2, self.world_height-house.height-1))
            self.houses.append(house)
            self.add(house, z=0)                 #Add the house to the layer

    def generate_footsteps(self, quantity):
        #print "Generating %d footsteps..." % quantity
        for i in range(quantity):
            pisadas = cocos.sprite.Sprite('pisadas%d.png' % random.randint(1, 4))
            pisadas.position = (random.randint(0 + pisadas.width/2, self.world_width-pisadas.width-1), random.randint(0 + pisadas.height/2, self.world_height-pisadas.height-1))
            self.add(pisadas, z=0)                 #Add the house to the layer

    def generate_vines(self, quantity):
        #print "Generating %d vines..." % quantity
        for i in range(quantity):
            vine = cocos.sprite.Sprite('vine.png')
            vine.position = (random.randint(0 + vine.width/2, self.world_width-vine.width-1), random.randint(0 + vine.height/2, self.world_height-vine.height-1))
            self.add(vine, z=0)                 #Add the house to the layer

    def generate_pines(self, quantity):
        #print "Generating %d pines..." % quantity
        for i in range(quantity):
            pino = cocos.sprite.Sprite('pino.png')
            pino.position = (random.randint(0 + pino.width/2, self.world_width-pino.width-1), random.randint(0 + pino.height/2, self.world_height-pino.height-1))
            pino.scale = 0.4
            self.add(pino, z=0)                 #Add the house to the layer

    def generate_rocks(self, quantity):
        #print "Generating %d rocks..." % quantity
        for i in range(quantity):
            obj = cocos.sprite.Sprite(random.choice(['roca1.png', 'roca2.png']))
            obj.position = (random.randint(0 + obj.width/2, self.world_width-obj.width-1), random.randint(0 + obj.height/2, self.world_height-obj.height-1))
            self.add(obj, z=0)                 #Add the house to the layer

    def on_enter(self):
        super(TerrainLayer, self).on_enter()
        self.scroller = self.get_ancestor(cocos.layer.ScrollingManager)
        self.scene = self.get_ancestor(cocos.scene.Scene)

    def on_key_press(self, kp, modifiers):
        self.pressed_keys.add(kp)

        #Toggle music ON/OFF
        if (kp == key.M) and (pyglet.media.have_avbin):
            if audio_player.playing:
                audio_player.pause()
            else:
                audio_player.play()


    def on_key_release(self, kp, modifiers):
        #use discard() method instead of remove() to avoid possible KeyError exception
        self.pressed_keys.discard(kp)


    def step(self, dt):
        #Update the position of the hero based on the pressed keys,
        #and set the proper direction
        if key.LEFT in self.pressed_keys:
            self.player.position = (self.player.position[0] - self.player.speed * dt, self.player.position[1])
            self.player.set_direction('left', dt)
        if key.RIGHT in self.pressed_keys:
            self.player.position = (self.player.position[0] + self.player.speed * dt, self.player.position[1])
            self.player.set_direction('right', dt)
        if key.UP in self.pressed_keys:
            self.player.position = (self.player.position[0], self.player.position[1] + self.player.speed * dt)
            self.player.set_direction('up', dt)
        if key.DOWN in self.pressed_keys:
            self.player.position = (self.player.position[0], self.player.position[1] - self.player.speed * dt)
            self.player.set_direction('down', dt)

        #Restrict the hero position to a valid one
        self.player.position = self.clamp(self.player, self.player.position)
        self.player.shadow.position = (self.player.position[0], self.player.position[1] - self.player.height/2 - 2)

        #Center the camera on the hero
        try:
            self.scroller.set_focus(*self.clamp_camera(self.player.position))
        except Exception, e:
            print str(e)

        #Zooming is only available when playing on easy mode
        if self.difficulty == 'easy':
            #Handle with zooming
            if key.PLUS in self.pressed_keys:
                self.scroller.scale = self.clamp_scale(self.scroller.scale + 0.2 * dt)
            if key.MINUS in self.pressed_keys:
                self.scroller.scale = self.clamp_scale(self.scroller.scale - 0.2 * dt)
            if key.NUM_MULTIPLY in self.pressed_keys:
                #Restore to the original scale
                self.scroller.scale = self.clamp_scale(1.0)

        #The king is following the hero!
        if self.king.is_following:
            #Update the position of the king; put it behind the hero
            self.king.follow(self.player.position, self.player.direction, dt)
            #If the hero decides to release the king...
            if key.SPACE in self.pressed_keys:
                self.king.is_following = False
                self.king.unfollow()
            if self.king.collides(self.the_cross):
                cocos.director.director.replace( Scene( VictoryLayer(), BackMenu() ))

        #If the king still isn't following the hero...
        else:
            #If the hero and the king collide, make the king follow the hero
            if self.player.collides(self.king):
                self.king.is_following = True
                outta_dodge = Label("Let's get outta Dodge!", font_size=14, font_name='Arial Black', color=(0, 0, 0, 255), anchor_x='center', anchor_y='center')
                outta_dodge.position = (self.king.position[0], self.king.position[1] + self.king.height/2 + 5)
                self.add(outta_dodge, z=2)
                outta_dodge.do(FadeOut(4))

        for enemy in self.enemies:
            if hasattr(enemy, 'radar'):
                if enemy.radar.is_point_inside(self.player.position):
                    cocos.director.director.replace( Scene( DefeatLayer(), BackMenu() ))
                if enemy.radar.is_point_inside(self.king.position) and (self.king.is_following):
                    cocos.director.director.replace( Scene( DefeatLayer(), BackMenu() ))
            enemy.patrol(dt)

    #Check that the new position is inside the allowed boundaries
    def clamp(self, actor, new_pos):
        x, y = new_pos
        if x - actor.width/2 < 0:
            x = 0 + actor.width/2
        elif x + actor.width/2 > self.world_width:
            x = self.world_width - actor.width/2
        if y - actor.height/2 < 0:
            y = 0 + actor.height/2
        elif y + actor.height/2> self.world_height:
            y = self.world_height - actor.height/2
        return x, y

    #Check that the new scale is between acceptable values
    def clamp_scale(self, new_scale):
        if new_scale > 2.0:
            new_scale = 2.0
        elif new_scale < 0.2:
            new_scale = 0.2
        #print "[*] New scale: %s" % new_scale
        return new_scale

    """
    Check whether we should move the camera or not, to avoid showing black areas outside the map.
    I'm NOT moving the camera when it's too close to the borders, i.e when:
    * x < window_width/2
    * x > world_width - window_width/2
    * y < window_height/2
    * y > world_height - window_height/2
    """
    def clamp_camera(self, position):
        cam_pos_x, cam_pos_y = position
        if (cam_pos_x < self.window_width/2):
            cam_pos_x = self.window_width/2
        elif (cam_pos_x > self.world_width - self.window_width/2):
            cam_pos_x = self.world_width - self.window_width/2
        if (cam_pos_y < self.window_height/2):
             cam_pos_y = self.window_height/2
        elif (cam_pos_y > self.world_height - self.window_height/2):
             cam_pos_y = self.world_height - self.window_height/2
        return (cam_pos_x, cam_pos_y)

##################################################################################################

class MainMenu(Menu):
    """ Main menu for the game """

    def __init__( self ):
        # call superclass with the title
        super( MainMenu, self ).__init__("Himalia")

        #Font attributes for the menu title
        self.font_title['font_name'] = '4990810'
        self.font_title['font_size'] = 50
        self.font_title['color'] = (255,255,255,255)

        #Font attributes for the menu items
        self.font_item['font_name'] = '4990810'
        self.font_item['color'] = (144,84,48,255)
        self.font_item['font_size'] = 25

        #Font attributes for the selected menu item
        self.font_item_selected['font_name'] = '4990810'
        self.font_item_selected['color'] = (221,84,1,255)
        self.font_item_selected['font_size'] = 25
        
        # example: menus can be vertical aligned and horizontal aligned
        self.menu_valign = CENTER
        self.menu_halign = CENTER

        #Menu items                
        items = []
        items.append( MenuItem('New Game', self.on_new_game ) )
        items.append( MenuItem('Help', self.on_help ) )
        items.append( MenuItem('Credits', self.on_credits ) )
        items.append( MenuItem('Quit', self.on_quit ) )

        self.create_menu( items, zoom_in(), zoom_out() )

    #Callbacks
    def on_new_game(self):
        cocos.director.director.replace(JumpZoomTransition(Scene(DifficultyMenu()), duration=1))

    def on_help(self):
        color_layer = ColorLayer(17,34,99,255)
        cocos.director.director.replace(JumpZoomTransition(Scene(HelpLayer(), BackMenu()), duration=1))

    def on_quit(self):
        cocos.director.director.pop()

    def on_credits(self):
        color_layer = ColorLayer(17,34,99,255)
        #cocos.director.director.replace(JumpZoomTransition(Scene(CloudsLayer(), CreditsLayer(), BackMenu()), duration=1))
        cocos.director.director.replace(JumpZoomTransition(Scene(CreditsLayer(), BackMenu()), duration=1))


##################################################################################################

class BackMenu(Menu):
    def __init__(self):
        super( BackMenu, self ).__init__()

        #Font attributes for the menu items
        self.font_item['font_name'] = '4990810'
        self.font_item['color'] = (144,84,48,255)
        self.font_item['font_size'] = 25

        #Font attributes for the selected menu item
        self.font_item_selected['font_name'] = '4990810'
        self.font_item_selected['color'] = (221,84,1,255)
        self.font_item_selected['font_size'] = 25

        # example: menus can be vertical aligned and horizontal aligned
        self.menu_valign = BOTTOM
        self.menu_halign = CENTER

        #Menu items
        items = []
        items.append( MenuItem('Back to Main Menu', self.on_back ) )
        self.create_menu( items, zoom_in(), zoom_out() )

    def on_back(self):
        cocos.director.director.replace(JumpZoomTransition(Scene(MainMenu()), duration=1))

    def on_quit(self):
        self.on_back()

##################################################################################################

class DifficultyMenu(Menu):
    def __init__(self):
        super( DifficultyMenu, self ).__init__("Choose your level")

        #Font attributes for the menu title
        self.font_title['font_name'] = '4990810'
        self.font_title['font_size'] = 50
        self.font_title['color'] = (255,255,255,255)

        #Font attributes for the menu items
        self.font_item['font_name'] = '4990810'
        self.font_item['color'] = (144,84,48,255)
        self.font_item['font_size'] = 25

        #Font attributes for the selected menu item
        self.font_item_selected['font_name'] = '4990810'
        self.font_item_selected['color'] = (221,84,1,255)
        self.font_item_selected['font_size'] = 25

        # example: menus can be vertical aligned and horizontal aligned
        self.menu_valign = BOTTOM
        self.menu_halign = CENTER

        #Menu items
        items = []
        items.append( MenuItem('Easy, easy...', self.on_easy ) )
        items.append( MenuItem('Somewhat difficult', self.on_medium ) )
        items.append( MenuItem('Frustration for sure', self.on_hard ) )
        items.append( MenuItem('Back to Main Menu', self.on_back ) )
        self.create_menu( items, zoom_in(), zoom_out() )

    def on_easy(self):
        scroller = ScrollingManager()
        scroller.add(TerrainLayer('desert.png', 1280, 1280, 'easy'))
        cocos.director.director.replace(RotoZoomTransition(Scene(scroller), duration=1))

    def on_medium(self):
        scroller = ScrollingManager()
        scroller.add(TerrainLayer('desert.png', 1280, 1280, 'medium'))
        cocos.director.director.replace(RotoZoomTransition(Scene(scroller), duration=1))

    def on_hard(self):
        scroller = ScrollingManager()
        scroller.add(TerrainLayer('desert.png', 1280, 1280, 'hard'))
        cocos.director.director.replace(RotoZoomTransition(Scene(scroller), duration=1))

    def on_back(self):
        cocos.director.director.replace(JumpZoomTransition(Scene(MainMenu()), duration=1))

    def on_quit(self):
        self.on_back()

##################################################################################################
